"""
Computers

This module contains actions related to Active Directory Computers.
"""
import uuid
import functools
import logging

from typing import List, Optional, Union, Dict, Any

from ppa import ui

from active_directory import computers, groups

from .exceptions import NoComputerFound
from ._wrappers import ldap_function, process_computer
from .records import Computer, DomainController

logger = logging.getLogger(__name__)


@ldap_function
def get_all(domain_controller: DomainController, search_base: Optional[str] = None) -> List[Computer]:
    """Get all computers.

!!! danger "This operation can be slow on large domains"

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __search_base:__ a distinguishedName to use as the root of the search (defaults to the root of the domain)

??? success "Output"

    A list of [Computer Dictionaries](../inputs_outputs/#computer).

??? example
    ```YAML linenums="1"
    - active_directory.computers.get_all:
      load:
        domain_controller: domain_controller_info
      save: all_computers
    ```
"""
    return list(map(process_computer, computers.get_all(domain_controller, search_base=search_base)))


@ldap_function
def by_samaccountname(domain_controller: DomainController, sAMAccountName: str) -> Computer:
    """Get a computer by its 'sAMAccountName' attribute.

!!! question "Minimum Plugin Version: 1.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __sAMAccountName:__ The computer's `sAMAccountName`

??? success "Output"
    A single [Computer Dictionary](../inputs_outputs/#computer).

??? example
    ```YAML linenums="1"
    - active_directory.computers.by_samaccountname:
        sAMAccountName: example.computer
      load:
        domain_controller: domain_controller_info
      save: example_computer
    ```
"""
    try:
        return process_computer(computers.search(domain_controller, sAMAccountName=sAMAccountName)[0])
    except IndexError:
        raise NoComputerFound(f"No computer found with sAMAccountName '{sAMAccountName}'.")


@ldap_function
def by_distinguishedname(domain_controller: DomainController, distinguishedName: str) -> Computer:
    """Get a computer by its 'distinguishedName' attribute.

!!! question "Minimum Plugin Version: 1.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ The computer's `distinguishedName`

??? success "Output"
    A single [Computer Dictionary](../inputs_outputs/#computer).

??? example
    ```YAML linenums="1"
    - active_directory.computers.by_distinguishedname:
        distinguishedName: cn=Example Computer,OU=Computers,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller_info
      save: example_computer
    ```
"""
    try:
        return process_computer(computers.search(domain_controller, distinguishedName=distinguishedName)[0])
    except IndexError:
        raise NoComputerFound(f"No computer found with distinguishedName '{distinguishedName}'.")


@ldap_function
def search(
    domain_controller: DomainController, search_params: Dict[str, Any], search_base: Optional[str] = None
) -> List[Computer]:
    """Search for computers using LDAP attributes & values.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __search_params:__ a dictionary containing [computer keys](../inputs_outputs#computer) & values to use in the search

    - __search_base:__ a distinguishedName to use as the root of the search (defaults to the root of the domain)

??? success "Output"

    A list of [Computer Dictionaries](../inputs_outputs/#computer).

??? example

    Searching for all computers in the __Computers__ CN whose `cn` starts with __DC__:

    ```YAML linenums="1"
    - active_directory.computers.search:
        search_params:
          cn: DC*
        search_base: CN=Computers,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller_info
      save: domain_controllers
    ```

!!! tip

    You can use `*` as a wildcard at the end of search values.
"""
    return list(map(process_computer, computers.search(domain_controller, search_base=search_base, **search_params)))


@ldap_function
def add_to_group(domain_controller: DomainController, computer_distinguishedName: str, group_distinguishedName: str) -> None:
    """Add a computer to a group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __user_distinguishedName:__ the distinguishedName of the computer

    - __group_distinguishedName:__ the distinguishedName of the group

??? success "Output"
    Nothing is outputted by this action.

??? example
    Adding computer _Example Computer_ to the group _Example Group_.

    ```YAML linenums="1"
    - active_directory.computers.add_to_group:
        computer_distinguishedName: "CN=Example Computer,CN=Computers,DC=Example,DC=Domain"
        group_distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
"""
    groups.add_member(
        domain_controller, distinguished_name=group_distinguishedName, member_distinguished_name=computer_distinguishedName,
    )


@ldap_function
def remove_from_group(domain_controller: DomainController, computer_distinguishedName: str, group_distinguishedName: str) -> None:
    """Remove a computer from a group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __user_distinguishedName:__ the distinguishedName of the computer

    - __group_distinguishedName:__ the distinguishedName of the group

??? success "Output"
    Nothing is outputted by this action.

??? example
    Removing computer _Example Computer_ from the group _Example Group_.

    ```YAML linenums="1"
    - active_directory.computers.remove_from_group:
        computer_distinguishedName: "CN=Example Computer,CN=Computers,DC=Example,DC=Domain"
        group_distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    groups.remove_member(
        domain_controller, distinguished_name=group_distinguishedName, member_distinguished_name=computer_distinguishedName
    )


@ldap_function
def set_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: str) -> None:
    """Set a computer attribute value.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the computer

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to set

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.computers.set_attribute:
        distinguishedName: "CN=Example Computer,CN=Computers,DC=Example,DC=Domain"
        name: info
        value: Example Info Value
      load:
        domain_controller: domain_controller_info
    ```
    """
    return computers.set_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


@ldap_function
def replace_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: str) -> None:
    """Replace a computer attribute value.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the computer

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to set

??? success "Output"
    Nothing is outputted by this action.

??? example
    - active_directory.computers.replace_attribute:
        distinguishedName: "CN=Example Computer,CN=Computers,DC=Example,DC=Domain"
        name: info
        value: Example Info Value
      load:
        domain_controller: domain_controller_info
    ```
    """
    return computers.replace_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


@ldap_function
def delete_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: str) -> None:
    """Delete a certain value from a computer attribute.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"
    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the computer

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to delete

??? success "Output"
    Nothing is outputted by this action.

??? example
    - Clearing the existing `info` value of a computer

    - The search outputs a list, so the computer is accessed using an index in the second action

    ```YAML linenums="1"
    - active_directory.computers.search:
        sAMAccountName: Example Computer
      load:
        domain_controller: domain_controller_info
      save: computers

    - active_directory.computers.delete_attribute:
        name: info
      load:
        domain_controller: domain_controller_info
        distinguishedName: computers.0.distinguishedName
        value: computers.0.info
    ```
    """
    return computers.delete_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


@ldap_function
def clear_attribute(domain_controller: DomainController, distinguishedName: str, name: str) -> None:
    """Clear a particular LDAP attribute on a computer.

Works for single-valued & multi-valued string LDAP attributes.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the computer

    - __name:__ the name of the LDAP attribute

??? success "Output"

    Nothing is outputted by this action.

??? example

    Searching for a computer & clearing its __info__ field:

    ```YAML linenums="1"
    - active_directory.computers.by_samaccountname:
        sAMAccountName: IIS-001
      load:
        domain_controller: domain_controller
      save: computer

    - active_directory.computers.clear_attribute:
        name: info
      load:
        distinguishedName: computer.distinguishedName
        domain_controller: domain_controller
    ```
"""
    existing_value = by_distinguishedname(domain_controller, distinguishedName)[name]

    if existing_value is None:
        # Nothing to do as the attribute has no value.
        return

    delete_func = functools.partial(delete_attribute, domain_controller, distinguishedName, name=name)
    list(map(lambda v: delete_func(value=v), existing_value)) if isinstance(existing_value, list) else delete_func(
        value=existing_value
    )


@ldap_function
def exists(domain_controller: DomainController, **search_params: Union[str, int, None]) -> bool:
    """Search for computers using LDAP attributes & values to identify if any were found.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __search_params:__ a dictionary containing [computer keys](../inputs_outputs#computer) & values to use in the search


??? success "Output"

    A boolean is outputted by this action

    - `true` if one or more computers are found

    - `false` if no computers are found

??? example

    Searching using a unique attribute:

    ```YAML linenums="1"
    - active_directory.computers.exists:
        distinguishedName: "CN=Computer Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: item_exists
    ```

    Searching using a non-unique attribute:

    ```YAML linenums="1"
    - active_directory.computers.exists:
        cn: QA*
      load:
        domain_controller: domain_controller_info
      save: item_exists
    ```

!!! tip "Wildcard Searching"

    It is possible to use `*` as a wildcard at the __end__ of search values, but this can make the search slow.

    """
    return bool(computers.search(domain_controller, **search_params))


@ldap_function
def move(domain_controller: DomainController, distinguishedName: str, parent_distinguishedName: str) -> None:
    """Move a computer to a different OU or CN.

!!! question "Minimum Plugin Version: 1.5.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the computer to move

    - __parent_distinguishedName:__ the distinguishedName of the OU or CN to move the computer to

??? success "Output"

    Nothing is outputted by this action.

??? example

    Moving a computer from the 'Users' CN to the 'Staff Computers' OU:

    ```YAML linenums="1"
    - active_directory.computers.move:
        distinguishedName: "CN=Example Computer,CN=Users,DC=Example,DC=Domain"
        parent_distinguishedName: "OU=Staff Computers,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    return computers.move(domain_controller, distinguished_name=distinguishedName, new_parent=parent_distinguishedName)


def output_table(text: str, computer_list: List[Computer]) -> None:
    """Display a list of computers in a table.

The table will have the following columns:

- Common Name
- Distinguished Name
- Operating System
- OS Version
- Description

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"
    - __text:__ the title of the table

    - __computer_list:__ a list of [Computer Dictionaries](../inputs_outputs/#computer)

??? success "Output"
    Nothing is outputted by this action.

??? example
    - Getting all computers whose common names start with _Operations_

    - Saving the computers as a new variable called `operations_computers`

    - Using this action to show the `operations_computers` in a table

    ```YAML linenums="1"
    - active_directory.computers.search:
        cn: Operations
      load:
        domain_controller: domain_controller_info
      save: operations_computers

    - active_directory.computers.output_table:
        text: Operations Computers
      load:
        computer_list: operations_computers
    ```
    """
    ui.output_table(
        text,
        header=["Common Name", "Distinguished Name", "Operating System", "OS Version", "Description"],
        data=[
            [
                computer.get(field) or ""
                for field in ["cn", "distinguishedName", "operatingSystem", "operatingSystemVersion", "description"]
            ]
            for computer in computer_list
        ],
    )


def output_custom_table(text: str, header: List[str], attributes: List[str], computer_list: List[Computer]) -> None:
    """Display a list of computers in a table using custom headers & LDAP attributes.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"
    - __text:__ the title of the table

    - __header:__ a list of column names

    - __attributes:__ a list of LDAP attributes (one for each column name in the `header`)

    - __computer_list:__ a list of [Computer Dictionaries](../inputs_outputs/#computer) to display in the table

??? success "Output"
    Nothing is outputted by this action.

??? example
    Showing `cn`, `operatingSystem` & `objectSid` for computers in the variable `audit_computers`.

    ```YAML linenums="1"
    - active_directory.computers.output_custom_table:
        text: Custom Computers Table
        header:
          - Common Name
          - Operating System
          - SID
        attributes:
          - cn
          - operatingSystem
          - objectSid
      load:
        computer_list: audit_computers
    ```
    """
    ui.output_table(text, header=header, data=[[computer.get(attr) or "" for attr in attributes] for computer in computer_list])


def input_table(
    text: str, computer_list: List[Computer], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[Computer]:
    """Display a list of computers in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Common Name
- Distinguished Name
- Operating System
- Operating System Version

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"
    - __text:__ The title of the table.

    - __computer_list:__ A list of [Computer Dictionaries](../inputs_outputs/#computer) to display in the table

    - __minimum:__ The minimum number of acceptable selections.

    - __maximum:__ The maximum number of acceptable selections.

??? success "Output"
    A list of [Computer Dictionaries](../inputs_outputs/#computer).

??? tip
    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.
    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
      of selections.

??? example
    - Getting all computers whose common names start with _Test_

    - Saving them as a new variable called `test_computers`

    - Using this action to show the `test_computers` in a table, requiring the task operator selects at least 1

    - The selection is saved as a new variable called `selected_computers`

    ```YAML linenums="1"
    - active_directory.computers.search:
        cn: Test*
      load:
        domain_controller: domain_controller_info
      save: test_computers

    - active_directory.computers.input_table:
        text: "Please Select >= 1 Computer(s)"
        minimum: 1
      load:
        computer_list: test_computers
      save: selected_computers
    ```
"""
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=["Common Name", "Distinguished Name", "OS", "OS Version"],
            data=[
                [computer.get(field) or "" for field in ["cn", "distinguishedName", "operatingSystem", "operatingSystemVersion"]]
                for computer in computer_list
            ],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [computer_list[i] for i in indices]
