"""
Users

This module contains actions related to Active Directory Users.
"""
import logging
import uuid
import functools

from typing import List, Dict, Optional, Union, Any

from ppa import ui

from active_directory import helpers, users, groups, exceptions

from .exceptions import NoUserFound
from ._wrappers import ldap_function, idempotent_create, idempotent_remove, process_user, process_group
from .records import User, Group, DomainController


logger = logging.getLogger(__name__)


@ldap_function
def search(domain_controller: DomainController, search_params: Dict[str, Any], search_base: Optional[str] = None) -> List[User]:
    """Search for users using LDAP attributes & values.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __search_params:__ a dictionary containing [user keys](../inputs_outputs#user) & values to use in the search

    - __search_base:__ a distinguishedName to use as the root of the search (defaults to the root of the domain)

??? success "Output"

    A list of [User Dictionaries](../inputs_outputs/#user).

???+ warning "Search Speed"

    This operation can be slow on a large domain.

    Consider targeting the search using `search_base` where possible.

??? example

    Searching for all users in the __Users__ CN whose `sAMAccountName` starts with __admin__:

    ```YAML linenums="1"
    - active_directory.users.search:
        search_params:
          sAMAccountName: admin*
        search_base: CN=Users,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller_info
      save: users
    ```

!!! tip

    You can use `*` as a wildcard at the end of search values.
"""
    return list(map(process_user, users.search(domain_controller, search_base=search_base, **search_params)))


@ldap_function
def members_of(
    domain_controller: DomainController, group_distinguishedNames: List[str], search_base: Optional[str] = None
) -> List[User]:
    """Find users who are members of __any__ of the supplied groups.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __group_distinguishedNames:__ a list of group distinguishedNames

    - __search_base:__ a distinguishedName to use as the root of the search (defaults to the root of the domain)

??? success "Output"

    A list of [User Dictionaries](../inputs_outputs/#user).

???+ warning "Search Speed"

    This operation can be slow on a large domain, especially if multiple groups are supplied.

    Consider targeting the search using `search_base` where possible.

??? example "Single Group"

    Finding users in the __Remote Desktop Users__ group:

    ```YAML linenums="1" hl_lines="3"
    - active_directory.users.members_of:
        group_distinguishedNames:
          - CN=Remote Desktop Users,CN=Builtin,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller_info
      save: users
    ```

??? example "Targeted Search & Multiple Groups"

    Finding all users who are:

    - In the default __Users__ CN

    - In _either_ the __Account Operators__ _or_ __Remote Desktop Users__ groups

    ```YAML linenums="1" hl_lines="3 4 5"
    - active_directory.users.members_of:
        group_distinguishedNames:
          - CN=Remote Desktop Users,CN=Builtin,DC=Example,DC=Domain,DC=Com
          - CN=Account Operators,CN=Users,DC=Example,DC=Domain,DC=Com
        search_base: CN=Users,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller_info
      save: users
    ```
"""
    return list(
        map(
            process_user,
            users.members_of(domain_controller, group_distinguished_names=group_distinguishedNames, search_base=search_base),
        )
    )


@ldap_function
def not_members_of(
    domain_controller: DomainController, group_distinguishedNames: List[str], search_base: Optional[str] = None
) -> List[User]:
    """Find users who are not members of __any__ of the supplied groups.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __group_distinguishedNames:__ a list of group distinguishedNames

    - __search_base:__ a distinguishedName to use as the root of the search (defaults to the root of the domain)

??? success "Output"

    A list of [User Dictionaries](../inputs_outputs/#user).

???+ warning "Search Speed"

    This operation can be slow on a large domain, especially if multiple groups are supplied.

    Consider targeting the search using `search_base` where possible.

??? example "Single Group"

    Finding all users not in the __Remote Desktop Users__ group:

    ```YAML linenums="1" hl_lines="3"
    - active_directory.users.not_members_of:
        group_distinguishedNames:
          - CN=Remote Desktop Users,CN=Builtin,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller_info
      save: users
    ```

??? example "Targeted Search & Multiple Groups"

    Finding all users who are:

    - In the default __Users__ CN

    - Not in _either_ the __Account Operators__ _or_ __Remote Desktop Users__ groups

    ```YAML linenums="1" hl_lines="3 4 5"
    - active_directory.users.not_members_of:
        group_distinguishedNames:
          - CN=Remote Desktop Users,CN=Builtin,DC=Example,DC=Domain,DC=Com
          - CN=Account Operators,CN=Users,DC=Example,DC=Domain,DC=Com
        search_base: CN=Users,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller_info
      save: users
    ```
"""
    return list(
        map(
            process_user,
            users.not_members_of(domain_controller, group_distinguished_names=group_distinguishedNames, search_base=search_base),
        )
    )


@ldap_function
def by_samaccountname(domain_controller: DomainController, sAMAccountName: str) -> User:
    """Get a user by its 'sAMAccountName' attribute.

!!! question "Minimum Plugin Version: 1.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __sAMAccountName:__ The user's `sAMAccountName`

??? success "Output"
    A single [User Dictionary](../inputs_outputs/#user).

??? example
    ```YAML linenums="1"
    - active_directory.users.by_samaccountname:
        sAMAccountName: example.user
      load:
        domain_controller: domain_controller_info
      save: example_user
    ```
"""
    try:
        return process_user(users.search(domain_controller, sAMAccountName=sAMAccountName)[0])
    except IndexError:
        raise NoUserFound(f"No user found with sAMAccountName '{sAMAccountName}'.")


@ldap_function
def by_distinguishedname(domain_controller: DomainController, distinguishedName: str) -> User:
    """Get a user by its 'distinguishedName' attribute.

!!! question "Minimum Plugin Version: 1.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ The user's `distinguishedName`

??? success "Output"
    A single [User Dictionary](../inputs_outputs/#user).

??? example
    ```YAML linenums="1"
    - active_directory.users.by_distinguishedname:
        distinguishedName: cn=Example User,CN=Users,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller_info
      save: example_user
    ```
"""
    try:
        return process_user(users.search(domain_controller, distinguishedName=distinguishedName)[0])
    except IndexError:
        raise NoUserFound(f"No user found with distinguishedName '{distinguishedName}'.")


@ldap_function
def get_interactive(
    domain_controller: DomainController,
    title: Optional[str] = None,
    search_attribute: Optional[str] = "sAMAccountName",
    exclude: Optional[Dict[str, Union[str, int]]] = None,
    search_base: Optional[str] = None,
) -> User:
    """Start an interactive search for a user.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __title:__ a title displayed to the Task Operator (defaults to `Get User Account`)

    - __search_attribute:__ an attribute from the following list:

        - _sAMAccountName_ (default value)
        - _cn_
        - _mail_

    - __exclude:__ a dictionary containing LDAP attribute names & regular expressions to test their values

    - __search_base:__ a distinguishedName to use as the root of the search (defaults to the root of the domain)

??? success "Output"

    A single [User Dictionary](../inputs_outputs/#user).

!!! tip "Automatic Wildcards"

    When using this action all provided search terms will have a wildcard appended.

??? example

    - Interactively searching for a user inside the builtin __Users__ CN

    - All users whose `sAMAccountName` contains __admin__ are excluded from the results

    ```YAML linenums="1"
    - active_directory.users.get_interactive:
        search_base: CN=Users,DC=Example,DC=Domain,DC=Com
        exclude:
          sAMAccountName: .*admin.*
      load:
        domain_controller: domain_controller_info
      save: user
    ```
    """
    return process_user(
        helpers.get_user(
            domain_controller,
            search_attribute=search_attribute if search_attribute in {"sAMAccountName", "cn", "mail"} else "sAMAccountName",
            search_base=search_base,
            auto_wildcard=True,
            auto_select=False,
            exclude=exclude,
            title=title,
        )
    )


@ldap_function
@idempotent_create
def add_to_group(domain_controller: DomainController, user_distinguishedName: str, group_distinguishedName: str) -> None:
    """Add a user to a group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __user_distinguishedName:__ the distinguishedName of the user

    - __group_distinguishedName:__ the distinguishedName of the group

??? success "Output"
    Nothing is outputted by this action.

??? example
    Adding a user to a group.

    - The user & group are searched for using the _sAMAccountName_ attribute, & saved as new variables _user_ & _group_.

    - This action is then supplied the required _distinguishedName_ using fields from the _user_ & _group_ variables.
    ```YAML linenums="1"
    - active_directory.groups.search:
        sAMAccountName: example.group
      load:
        domain_controller: domain_controller_info
      save: group

    - active_directory.users.search:
        sAMAccountName: example.user
      load:
        domain_controller: domain_controller_info
      save: user

    - active_directory.users.add_to_group:
      load:
        user_distinguishedName: user.distinguishedName
        group_distinguishedName: group.distinguishedName
        domain_controller: domain_controller_info
    ```
"""
    groups.add_member(
        domain_controller, distinguished_name=group_distinguishedName, member_distinguished_name=user_distinguishedName
    )


@ldap_function
def add_to_groups_interactive(domain_controller: DomainController, distinguishedName: str) -> List[Group]:
    """Interactively add a user to one or more groups.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"
    A list of [Group Dictionaries](../inputs_outputs/#group) the user was added to.

??? example

    Triggering an interactive process where user _Example User_ can be added to one or more groups.

    ```YAML linenums="1"
    - active_directory.users.add_to_groups_interactive:
        distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    return list(map(process_group, helpers.add_user_to_groups(domain_controller, distinguished_name=distinguishedName)[0]))


@ldap_function
@idempotent_remove
def remove_from_group(domain_controller: DomainController, user_distinguishedName: str, group_distinguishedName: str) -> None:
    """Remove a user from a group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __user_distinguishedName:__ the distinguishedName of the user

    - __group_distinguishedName:__ The distinguishedName of the group

??? success "Output"
    Nothing is outputted by this action.

??? example
    Removing user _Example User_ from the group _Example Group_.
    ```YAML linenums="1"
    - active_directory.user.remove_from_group:
        user_distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
        group_distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    groups.remove_member(
        domain_controller, distinguished_name=group_distinguishedName, member_distinguished_name=user_distinguishedName
    )


@ldap_function
def remove_from_groups_interactive(domain_controller: DomainController, distinguishedName: str) -> List[Group]:
    """Interactively remove a user from one or more groups.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"
    A list of [Group Dictionaries](../inputs_outputs/#group) the user was removed from.

??? example
    Triggering an interactive process where user _Example User_ can be removed from one or more groups.
    ```YAML linenums="1"
    - active_directory.users.remove_from_groups_interactive:
        distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    return list(map(process_group, helpers.remove_user_from_groups(domain_controller, distinguished_name=distinguishedName)))


@ldap_function
@idempotent_create
def create(
    domain_controller: DomainController,
    distinguishedName: str,
    sAMAccountName: str,
    extra_params: Optional[Dict[str, Any]] = None,
) -> None:
    """Create a new user.

!!! question "Minimum Plugin Version: 1.1.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __sAMAccountName:__ the new user sAMAccountName

    - __distinguishedName:__ the new user distinguishedName

    - __extra_params:__ a dictionary containing any extra LDAP attributes & values for the user

??? success "Output"
    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - active_directory.users.create:
        sAMAccountName: example.user
        distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
        extra_params:
          info: An example user
          cn: Example User
      load:
        domain_controller: domain_controller_info
    ```
    """
    return users.create(
        domain_controller, distinguished_name=distinguishedName, samaccountname=sAMAccountName, **extra_params or {}
    )


@ldap_function
@idempotent_remove
def delete(domain_controller: DomainController, distinguishedName: str) -> None:
    """Delete a user.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user to delete

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.delete:
        distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    return users.delete(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def move(domain_controller: DomainController, distinguishedName: str, parent_distinguishedName: str) -> None:
    """Move a user to a different OU or CN.

!!! question "Minimum Plugin Version: 1.5.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user to move

    - __parent_distinguishedName:__ the distinguishedName of the OU or CN to move the user to

??? success "Output"

    Nothing is outputted by this action.

??? example

    Moving a user from the 'Users' CN to the 'Staff' OU:

    ```YAML linenums="1"
    - active_directory.users.move:
        distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
        parent_distinguishedName: "OU=Staff,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    return users.move(domain_controller, distinguished_name=distinguishedName, new_parent=parent_distinguishedName)


@ldap_function
def set_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: Union[str, int]) -> None:
    """Set a user attribute value.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to set

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.get_interactive:
      load:
        domain_controller: domain_controller_info
      save: user

    - active_directory.users.set_attribute:
        name: info
        value: Example Info Value
      load:
        distinguishedName: user.distinguishedName
        domain_controller: domain_controller_info
    ```
    """
    return users.set_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


@ldap_function
def replace_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: Union[str, int]) -> None:
    """Replace a user attribute value.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to set

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.get_interactive:
      load:
        domain_controller: domain_controller_info
      save: user

    - active_directory.users.replace_attribute:
        name: info
        value: Example Info Value
      load:
        distinguishedName: user.distinguishedName
        domain_controller: domain_controller_info
    ```
    """
    return users.replace_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


@ldap_function
def delete_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: Union[str, int]) -> None:
    """Delete a certain value from a user attribute.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to delete

??? success "Output"
    Nothing is outputted by this action.

??? example

    Clearing the existing _info_ value of a user.

    ```YAML linenums="1"
    - active_directory.users.get_interactive:
      load:
        domain_controller: domain_controller_info
      save: user

    - active_directory.users.delete_attribute:
        name: info
      load:
        distinguishedName: user.distinguishedName
        value: user.info
        domain_controller: domain_controller_info
    ```
    """
    return users.delete_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


@ldap_function
def clear_attribute(domain_controller: DomainController, distinguishedName: str, name: str) -> None:
    """Clear a particular LDAP attribute on a user.

Works for single-valued & multi-valued string LDAP attributes.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __name:__ the name of the LDAP attribute

??? success "Output"

    Nothing is outputted by this action.

??? example

    Searching for a user & clearing their __carLicense__ field:

    ```YAML linenums="1"
    - active_directory.users.by_samaccountname:
        sAMAccountName: john.smith
      load:
        domain_controller: domain_controller
      save: user

    - active_directory.users.clear_attribute:
        name: carLicense
      load:
        distinguishedName: user.distinguishedName
        domain_controller: domain_controller
    ```
"""
    existing_value = by_distinguishedname(domain_controller, distinguishedName)[name]

    if existing_value is None:
        # Nothing to do as the attribute has no value.
        return

    delete_func = functools.partial(delete_attribute, domain_controller, distinguishedName, name=name)
    list(map(lambda v: delete_func(value=v), existing_value)) if isinstance(existing_value, list) else delete_func(
        value=existing_value
    )


@ldap_function
def set_password(domain_controller: DomainController, distinguishedName: str, password: str) -> None:
    """Set a user's password.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __password:__ the new password

??? success "Output"

    Nothing is outputted by this action.

??? example
    - Setting the password for user _John Smith_ in the root _Users_ CN

    - The PPA UI _input_password_ action is used to get & save the new password

    ```YAML linenums="1"
    - ppa.ui.input_password:
        text: New Password
      save: new_password

    - active_directory.users.set_password:
        distinguishedName: "CN=John Smith,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
        password: new_password
    ```
"""
    return users.set_password(domain_controller, distinguished_name=distinguishedName, new_password=password)


def set_password_interactive(domain_controller: DomainController, distinguishedName: str) -> None:
    """Interactively set a user's password. This operation:

- Asks the task operator to supply & confirm a new password.
- Attempts to set the password for the user.
- If the password is refused by Active Directory the operation will be repeated repeated.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary
    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.set_password_interactive:
        distinguishedName: "CN=John Smith,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
"""

    def _get_password():
        while True:
            password = ui.input_password("New Password", meter=True, element_id="new_password")
            if password == ui.input_password("Confirm Password", meter=True, element_id="confirm_password"):
                break
            else:
                ui.output_error("Passwords must match.", element_id="password_error")
        return password

    def _set_password(password):
        try:
            set_password(domain_controller, distinguishedName, password)
            return True
        except exceptions.InvalidPassword as e:
            ui.output_error(str(e), element_id="invalid_password_exception")
            ui.output_error("Please try again.", element_id="invalid_password_error")
            return False

    for _ in range(0, 10):
        if _set_password(_get_password()):
            return
    raise Exception("Failed to set password after 10 attempts. Operation failed.")


@ldap_function
def disable(domain_controller: DomainController, distinguishedName: str) -> None:
    """Disable a user.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.disable:
        distinguishedName: "CN=John Smith,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
"""
    return users.disable(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def enable(domain_controller: DomainController, distinguishedName: str) -> None:
    """Enable a user.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.enable:
        distinguishedName: "CN=John Smith,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
"""
    return users.enable(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def unlock(domain_controller: DomainController, distinguishedName: str) -> None:
    """Unlock a user.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.unlock:
        distinguishedName: "CN=John Smith,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
"""
    return users.unlock(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def force_password_change(domain_controller: DomainController, distinguishedName: str) -> None:
    """Set the 'force password change at next logon' flag against a user.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.force_password_change:
        distinguishedName: "CN=John Smith,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
"""
    return users.force_password_change(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def get_all(domain_controller: DomainController, search_base: Optional[str] = None) -> List[User]:
    """Get all users.

!!! danger "This operation can be slow on large domains"

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __search_base:__ a distinguishedName to use as the root of the search (defaults to the root of the domain)

??? success "Output"

    A list of [User Dictionaries](../inputs_outputs/#user).

??? example

    ```YAML linenums="1"
    - active_directory.users.get_all:
      load:
        domain_controller: domain_controller_info
      save: all_users
    ```
"""
    return list(map(process_user, users.get_all(domain_controller, search_base=search_base)))


@ldap_function
def group_memberships(domain_controller: DomainController, distinguishedName: str, nested: Optional[bool] = False) -> List[Group]:
    """Get the group memberships of a user.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __nested:__ Set to `true` to include nested group memberships in the search

??? success "Output"
    A list of group dictionaries.

??? warning
    Getting nested group memberships can be slow.

??? example
    - Getting all group memberships for user _Example User_

    - Saving the results as a new variable called `group_memberships`

    ```YAML linenums="1"
    - active_directory.users.group_memberships:
        distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
        nested: true
      load:
        domain_controller: domain_controller_info
      save: group_memberships
    ```
    """
    return list(
        map(process_group, users.group_memberships(domain_controller, distinguished_name=distinguishedName, nested=nested))
    )


def test_credentials(address: str, port: str, domain: str, username: str, password: str) -> None:
    """Test a user's credentials. Often used after resetting a password.

Takes the same inputs a [DomainController](../inputs_outputs/#domaincontroller) dictionary contains, but each input is supplied separately.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __address__: the Domain Controller IP or DNS address

    - __domain__: the FQDN of the Active Directory domain

    - __username__: username for authentication

    - __password__: password for authentication

??? success "Output"
    Nothing is outputted by this action.

??? example
    Testing a new password for user _John Smith_.

    - The `new_password` variable was set in an earlier step (not shown)

    - Note the `domain_controller` dictionary is used to supply the `address`, `port`, & `domain` keys

    ```YAML linenums="1"
    - active_directory.users.test_credentials:
        username: john.smith
      load:
        address: domain_controller.address
        port: domain_controller.port
        domain: domain_controller.domain
        password: new_password
    ```
"""
    return users.test_credentials(address=address, port=port, domain=domain, username=username, password=password)


def output_table(text: str, user_list: List[User]) -> None:
    """Display a list of users in a table. The table will have the following columns:

The table will have the following columns:

- Common Name
- SAM Account Name
- Email Address

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __user_list:__ a list of [User Dictionaries](../inputs_outputs/#user) to display in the table

??? success "Output"
    Nothing is outputted by this action.

??? example
    - Getting all users whose common names start with _Test_

    - Saving the results as a new variable called `test_users`

    - Using this action to show the `test_users` in a table

    ```YAML linenums="1"
    - active_directory.users.search:
        cn: Test*
      load:
        domain_controller: domain_controller_info
      save: test_users

    - active_directory.users.output_table:
        text: Test User Accounts
      load:
        user_list: test_users
    ```
    """
    ui.output_table(
        text,
        header=["Common Name", "SAM Account Name", "Email Address"],
        data=[[user.get(field) or "" for field in ["cn", "sAMAccountName", "mail"]] for user in user_list],
    )


def output_custom_table(text: str, header: List[str], attributes: List[str], user_list: List[User]) -> None:
    """Display a list of users in a table using custom headers & LDAP attributes.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __header:__ a list of column names

    - __attributes:__ a list of LDAP attributes (one for each column name in the `header`)

    - __user_list:__ a list of [User Dictionaries](../inputs_outputs/#user) to display in the table

??? success "Output"
    Nothing is outputted by this action.

??? example
    Showing `mail`, `distinguishedName`, & `userAccountControl` for users in the variable `test_users`.

    ```YAML linenums="1"
    - active_directory.users.output_custom_table:
        text: Custom Users Table
        header:
          - Email Address
          - Distinguished Name
          - User Account Control
        attributes:
          - mail
          - distinguishedName
          - userAccountControl
      load:
        user_list: test_users
    ```
    """
    ui.output_table(text, header=header, data=[[user.get(attr) or "" for attr in attributes] for user in user_list])


def input_table(text: str, user_list: List[User], minimum: Optional[int] = None, maximum: Optional[int] = None) -> List[User]:
    """Display a list of users in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Common Name
- SAM Account Name
- Email Address

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"
    - __text:__ The title of the table

    - __user_list:__ A list of [User Dictionaries](../inputs_outputs/#user) to display in the table

    - __minimum:__ The minimum number of acceptable selections.

    - __maximum:__ The maximum number of acceptable selections.

??? success "Output"
    A list of [User Dictionaries](../inputs_outputs/#user).

??? tip
    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.
    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
        of selections.

??? example
    - Getting all users whose common names start with _Test_

    - Saving the results as a new variable called `test_users`

    - Using this action to show the `test_users` in a table, requiring the task operator selects at least 1

    - The selection is saved as a new variable called `selected_users`

    ```YAML linenums="1"
    - active_directory.users.search:
        cn: Test*
      load:
        domain_controller: domain_controller_info
      save: test_users

    - active_directory.users.input_table:
        text: "Please Select >= 1 User(s)"
        minimum: 1
      load:
        user_list: test_users
      save: selected_users
    ```
    """
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=["Common Name", "SAM Account Name", "Email Address"],
            data=[[user.get(field) or "" for field in ["cn", "sAMAccountName", "mail"]] for user in user_list],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [user_list[i] for i in indices]


@ldap_function
def exists(domain_controller: DomainController, **search_params: Union[str, int, None]) -> bool:
    """Search for users using LDAP attributes & values to identify if any were found.

!!! question "Minimum Plugin Version: 1.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __search_params:__ a dictionary containing [user keys](../inputs_outputs#user) & values to use in the search


??? success "Output"

    A boolean is outputted by this action

    - `true` if one or more users are found

    - `false` if no users are found


??? example

    Searching using a unique attribute:

    ```YAML linenums="1"
    - active_directory.users.exists:
        distinguishedName: "CN=Full Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: item_exists
    ```

    Searching using a non-unique attribute:
    ```YAML linenums="1"
    - active_directory.users.exists:
        cn: "Maddison*"
      load:
        domain_controller: domain_controller_info
      save: item_exists
    ```

!!! tip "Wildcard Searching"

    It is possible to use `*` as a wildcard at the __end__ of search values, but this can make the search slow.

    """
    return bool(users.search(domain_controller, **search_params))


@ldap_function
def has_reversible_password(domain_controller: DomainController, distinguishedName: str) -> bool:
    """Determine if the supplied user's password is stored using reversible encryption.

!!! question "Minimum Plugin Version: 4.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user


??? success "Output"

    A boolean is outputted by this action

    - `true` if the password is stored using reversible encryption

    - `false` if the password is _not_ stored using reversible encryption


??? example

    ```YAML linenums="1"
    - active_directory.users.has_reversible_password:
        distinguishedName: "CN=Full Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: reversible
    ```
    """
    return users.has_reversible_password(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def smart_card_required(domain_controller: DomainController, distinguishedName: str) -> bool:
    """Determine if the supplied user account requires a smart card for interactive login.

!!! question "Minimum Plugin Version: 4.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user


??? success "Output"

    A boolean is outputted by this action

    - `true` if a smart card is required

    - `false` if a smart card is _not_ required


??? example

    ```YAML linenums="1"
    - active_directory.users.smart_card_required:
        distinguishedName: "CN=Full Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: card_required
    ```
    """
    return users.smart_card_required(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def password_will_expire(domain_controller: DomainController, distinguishedName: str) -> bool:
    """Determine if the supplied user account's password has an expiry.

!!! question "Minimum Plugin Version: 4.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user


??? success "Output"

    A boolean is outputted by this action

    - `true` if the user's password will expire

    - `false` if the user's password will _not_ expire


??? example

    ```YAML linenums="1"
    - active_directory.users.password_will_expire:
        distinguishedName: "CN=Full Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: password_will_expire
    ```
    """
    return users.password_will_expire(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def password_has_expired(domain_controller: DomainController, distinguishedName: str) -> bool:
    """Determine if the supplied user account's password has expired.

!!! question "Minimum Plugin Version: 4.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user


??? success "Output"

    A boolean is outputted by this action

    - `true` if the user's password has expired

    - `false` if the user's password has _not_ expired


??? example

    ```YAML linenums="1"
    - active_directory.users.password_has_expired:
        distinguishedName: "CN=Full Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: password_expired
    ```
    """
    return users.password_has_expired(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def password_not_required(domain_controller: DomainController, distinguishedName: str) -> bool:
    """Determine if the supplied user account can have a blank password.

!!! question "Minimum Plugin Version: 4.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user


??? success "Output"

    A boolean is outputted by this action

    - `true` if the user's password can be blank

    - `false` if the user's password _cannot_ be blank


??? example

    ```YAML linenums="1"
    - active_directory.users.password_not_required:
        distinguishedName: "CN=Full Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: password_not_required
    ```
    """
    return users.password_not_required(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def can_be_delegated(domain_controller: DomainController, distinguishedName: str) -> bool:
    """Determine if the supplied user account __has not__ been flagged as 'sensitive and not for delegation'.

!!! question "Minimum Plugin Version: 4.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user


??? success "Output"

    A boolean is outputted by this action

    - `true` if the user can be delegated

    - `false` if the user cannot be delegated


??? example

    ```YAML linenums="1"
    - active_directory.users.can_be_delegated:
        distinguishedName: "CN=Full Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: can_be_delegated
    ```
    """
    return users.can_be_delegated(domain_controller, distinguished_name=distinguishedName)
